#include <sys/stat.h>

#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "request.h"
#include "keyvalue.h"
#include "log.h"

inline int light_isdigit(int c) {
	return (c >= '0' && c <= '9');
}

inline int light_isalpha(int c) {
	return (c >= 'a' && c <= 'z') || 
		(c >= 'A' && c <= 'Z');
}

inline int light_isalnum(int c) {
	return light_isdigit(c) || light_isalpha(c);
}

int request_check_hostname(server *srv, connection *con, buffer *host) {
	enum { DOMAINLABEL, TOPLABEL } stage = TOPLABEL;
	size_t i;
	int label_len = 0;
	size_t host_len;
	char *colon;
	int is_ip = -1; /* -1 don't know yet, 0 no, 1 yes */ 
	int level = 0;
	
	/*
	 *       hostport      = host [ ":" port ]
	 *       host          = hostname | IPv4address
	 *       hostname      = *( domainlabel "." ) toplabel [ "." ]
	 *       domainlabel   = alphanum | alphanum *( alphanum | "-" ) alphanum
	 *       toplabel      = alpha | alpha *( alphanum | "-" ) alphanum
	 *       IPv4address   = 1*digit "." 1*digit "." 1*digit "." 1*digit
	 *       port          = *digit
	 */
	
	/* no Host: */
	if (!host || host->used == 0) return 0;
	
	host_len = host->used - 1;
	
	if (NULL != (colon = memchr(host->ptr, ':', host_len))) {
		register char *c = colon + 1;
		
		/* check portnumber */
		for (; *c; c++) {
			if (!light_isdigit(*c)) return -1;
		}
		
		/* remove the port from the host-len */
		host_len = colon - host->ptr;
	}
	
	/* Host is empty */
	if (host_len == 0) return -1;
	
	/* scan from the right and skip the \0 */
	for (i = host_len - 1; i + 1 > 0; i--) {
		char c = host->ptr[i];

		switch (stage) {
		case TOPLABEL: 
			if (c == '.') {
				/* only switch stage, if this is not the last character */
				if (i != host_len - 1) {
					if (label_len == 0) {
						return -1;
					}
					
					/* check the first character at right of the dot */
					if (is_ip == 0) {
						if (!light_isalpha(host->ptr[i+1])) {
							return -1; 
						}
					} else if (!light_isdigit(host->ptr[i+1])) {
						is_ip = 0;
					} else if ('-' == host->ptr[i+1]) {
						return -1;
					} else {
						/* just digits */
						is_ip = 1;
					}
						
					stage = DOMAINLABEL;
					
					label_len = 0;
					level++;
				} else if (i == 0) {
					/* just a dot and nothing else is evil */
					return -1;
				}
			} else if (i == 0) {
				/* the first character of the hostname */
				if (!light_isalpha(c)) {
					return -1;
				}
				label_len++;
			} else {
				if (c != '-' && !light_isalnum(c)) {
					return -1;
				}
				if (is_ip == -1) {
					if (!light_isdigit(c)) is_ip = 0;
				}
				label_len++;
			}
			
			break;
		case DOMAINLABEL:
			if (is_ip == 1) {
				if (c == '.') {
					if (label_len == 0) {
						return -1;
					}
					
					label_len = 0;
					level++;
				} else if (!light_isdigit(c)) {
					return -1;
				} else {
					label_len++;
				}
			} else {
				if (c == '.') {
					if (label_len == 0) {
						return -1;
					}
					
					/* c is either - or alphanum here */
					if ('-' == host->ptr[i+1]) {
						return -1;
					}
					
					label_len = 0;
					level++;
				} else if (i == 0) {
					if (!light_isalnum(c)) {
						return -1;
					}
					label_len++;
				} else {
					if (c != '-' && !light_isalnum(c)) {
						return -1;
					}
					label_len++;
				}
			}
			
			break;
		}
	}
	
	/* a IP has to consist of 4 parts */
	if (is_ip == 1 && level != 3) {
		return -1;
	}
	
	if (label_len == 0) {
		return -1;
	}
	
	return 0;
}

void dump_packet(const unsigned char *data, size_t len) {
	size_t i, j;
	
	if (len == 0) return;
	
	fprintf(stderr, "  ");
	for (i = 0; i < len; i++) {
		fprintf(stderr, "%02x ", data[i]);
		
		if ((i + 1) % 16 == 0) {
			fprintf(stderr, "  ");
			for (j = 0; j <= i % 16; j++) {
				unsigned char c;
				
				if (i-15+j >= len) break;
				
				c = data[i-15+j];
				
				fprintf(stderr, "%c", c > 32 && c < 128 ? c : '.');
			}
			
			fprintf(stderr, "\n  ");
		}
	}
	
	for (j = i % 16; j < 16; j++) {
		fprintf(stderr, "   ");
	}
	
	fprintf(stderr, "  ");
	for (j = i & ~0xf; j < len; j++) {
		unsigned char c;
		
		c = data[j];
		fprintf(stderr, "%c", c > 32 && c < 128 ? c : '.');
	}
	fprintf(stderr, "\n");
}


int http_request_parse(server *srv, connection *con) {
	size_t req_len;
	char *uri = NULL, *proto = NULL, *method = NULL, con_length_set, keep_alive_set = 0;
	int is_key = 1, key_len = 0, is_ws_after_key = 0, in_folding;
	char *value = NULL, *key = NULL;
	
	int line = 0;
	
	int request_line_stage = 0;
	size_t i, first;
	
	int done = 0;
	
	data_string *ds;
	
	/* 
	 * Request: "^(GET|POST|HEAD) ([^ ]+(\\?[^ ]+|)) (HTTP/1\\.[01])$" 
	 * Option : "^([-a-zA-Z]+): (.+)$"                    
	 * End    : "^$"
	 */
#if 0
	log_error_write(srv, __FILE__, __LINE__, "dbd", 
			con->fd, con->request.request, con->request.request->used);
	
	dump_packet(con->request.request->ptr, con->request.request->used);
#endif
	/* fill the local request buffer */
	req_len = con->request.request->used - 1;
	
	buffer_copy_string_buffer(con->parse_request, con->request.request);
	
	/* \r\n -> \n\0 */
	
	keep_alive_set = 0;
	con_length_set = 0;

	for (i = 0, first = 0; i < con->parse_request->used && line == 0; i++) {
		char *cur = con->parse_request->ptr + i;
		
		switch(*cur) {
		case '\r': 
			if (con->parse_request->ptr[i+1] == '\n') {
				http_method_t r;
				char *nuri = NULL;
				
				con->parse_request->ptr[i] = '\0';
				con->parse_request->ptr[i+1] = '\0';
				
				buffer_copy_string_len(con->request.request_line, con->parse_request->ptr, i);
				
				if (request_line_stage != 2) {
					con->http_status = 400;
					con->response.keep_alive = FALSE;
					
					return 0;
				}
				
				proto = con->parse_request->ptr + first;
				
				*(uri - 1) = '\0';
				*(proto - 1) = '\0';
				
				/* we got the first one :) */
				if (-1 == (r = get_http_method_key(method))) {
					con->http_status = 501;
					con->response.keep_alive = FALSE;
					
					log_error_write(srv, __FILE__, __LINE__, "sb", 
							"unknown http-method", con->request.request);
					
					return 0;
				}
				
				con->request.http_method = r;
				
				if (0 == strncmp(proto, "HTTP/1.", 7)) {
					if (proto[7] == '1') {
						con->request.http_version = HTTP_VERSION_1_1;
					} else if (proto[7] == '0') {
						con->request.http_version = HTTP_VERSION_1_0;
					} else { 
						con->http_status = 505;
						return 0;
					}
				} else {
					con->http_status = 400;
					log_error_write(srv, __FILE__, __LINE__, "");
					return 0;
				}
				
				if (0 == strncmp(uri, "http://", 7) &&
				    NULL != (nuri = strchr(uri + 7, '/'))) {
					/* ignore the host-part */
					
					buffer_copy_string_len(con->request.uri, nuri, proto - nuri - 1);
				} else {
					/* everything looks good so far */
					buffer_copy_string_len(con->request.uri, uri, proto - uri - 1);
				}
				
				con->http_status = 200;
				
				i++;
				line++;
				first = i+1;
			}
			break;
		case ' ':
			switch(request_line_stage) {
			case 0: 
				method = con->parse_request->ptr + first; 
				first = i + 1;
				break;
			case 1:
				uri = con->parse_request->ptr + first; 
				first = i + 1;
				break;
			default:
				/* ERROR */
				con->http_status = 400;
				con->response.keep_alive = FALSE;
				
				return 0;
			}
			
			request_line_stage++;
			break;
		}
	}
	
	in_folding = 0;
	
	for (; i < con->parse_request->used && !done; i++) {
		char *cur = con->parse_request->ptr + i;
		
		if (is_key) {
			/**
			 * 1*<any CHAR except CTLs or separators>
			 * CTLs == 0-31 + 127
			 * 
			 */
			switch(*cur) {
			case ':':
				is_key = 0;
				
				value = cur + 1;
				
				if (is_ws_after_key == 0) {
					key_len = i - first;
				}
				is_ws_after_key = 0;
					
				break;
			case '(':
			case ')':
			case '<':
			case '>':
			case '@':
			case ',':
			case ';':
			case '\\':
			case '\"':
			case '/':
			case '[':
			case ']':
			case '?':
			case '=':
			case '{':
			case '}':
				con->http_status = 400;
				con->response.keep_alive = FALSE;
				
				log_error_write(srv, __FILE__, __LINE__, "sbsd", 
						"invalid character in key", con->request.request, cur, *cur);
				return 0;
			case ' ':
			case '\t':
				if (i == first) {
					log_error_write(srv, __FILE__, __LINE__, "s", "(debug) Line Folding");
					
					is_key = 0;
					in_folding = 1;
					value = cur;
					
					break;
				}
				
				
				if (is_ws_after_key == 0) {
					key_len = i - first;
				}
				is_ws_after_key = 1;
				
				break;
			case '\r':
				if (con->parse_request->ptr[i+1] == '\n' && i == first) {
					/* End of Header */
					con->parse_request->ptr[i] = '\0';
					con->parse_request->ptr[i+1] = '\0';
					
					i++;
					
					done = 1;
					
					break;
				} else {
					log_error_write(srv, __FILE__, __LINE__, "sb", 
							"CR without LF", con->request.request);
					
					con->http_status = 400;
					con->response.keep_alive = FALSE;
					return 0;
				}
				/* fall thru */
			default:
				if (*cur < 32 || *cur == 127) {
					/* 400 */
					
					con->http_status = 400;
					con->response.keep_alive = FALSE;
					
					log_error_write(srv, __FILE__, __LINE__, "sbsd", 
							"CTL character in key", con->request.request, cur, *cur);
					
					return 0;
				}
				
				if (is_ws_after_key) {
					/* 400 */
					
					log_error_write(srv, __FILE__, __LINE__, "sb", 
							"WS character in key", con->request.request);
					
					con->http_status = 400;
					con->response.keep_alive = FALSE;
					return 0;
				}
				
				break;
			}
		} else {
			switch(*cur) {
			case '\r': 
				if (con->parse_request->ptr[i+1] == '\n') {
					/* End of Headerline */
					con->parse_request->ptr[i] = '\0';
					con->parse_request->ptr[i+1] = '\0';
					
					if (in_folding) {
						buffer_append_string(ds->value, value);
					} else {
						key = con->parse_request->ptr + first;
					
						if (NULL == (ds = (data_string *)array_get_unused_element(con->request.headers, TYPE_STRING))) {
							ds = data_string_init();
						}
						buffer_copy_string_len(ds->key, key, key_len);
						buffer_copy_string(ds->value, value);
						
						array_insert_unique(con->request.headers, (data_unset *)ds);
					}
					
					i++;
					first = i+1;
					is_key = 1;
					value = 0;
					key_len = 0;
					in_folding = 0;
				} else {
					log_error_write(srv, __FILE__, __LINE__, "sb", 
							"CR without LF", con->request.request);
					
					con->http_status = 400;
					con->response.keep_alive = FALSE;
					return 0;
				}
				break;
			case ' ':
			case '\t':
				/* strip leading WS */
				if (value == cur) value = cur+1;
			default:
				break;
			}
		}
	}
	
	con->header_len = i;
	
	/* retreive values */
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Host"))) {
		con->request.http_host = ds->value;
	} else {
		con->request.http_host = NULL;
	}
	
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Range"))) {
		/* bytes=.*-.* */
		
		if (0 == strncasecmp(ds->value->ptr, "bytes=", 6) &&
		    NULL != strchr(ds->value->ptr+6, '-')) {
			
			con->request.http_range = ds->value->ptr + 6;
		}
	}
	
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Connection"))) {
		con->keep_alive = (NULL != strstr(ds->value->ptr, "Keep-Alive")) ? TRUE : FALSE;
		keep_alive_set = 1;
	}
	
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Content-Type"))) {
		con->request.http_content_type = ds->value->ptr;
	}
	
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "If-None-Match"))) {
		con->request.http_if_none_match = ds->value->ptr;
	}
	
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "If-Modified-Since"))) {
		con->request.http_if_modified_since = ds->value->ptr;
	}
	
	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Content-Length"))) {
		char *err;
		con->request.content_length = strtol(ds->value->ptr, &err, 10);
		
		if (*err == '\0') {
			con_length_set = 1;
		} else {
			log_error_write(srv, __FILE__, __LINE__, "sb", "content-length broken:", ds->value);
		}
	}
	
	/* do some post-processing */

	if (con->request.http_version == HTTP_VERSION_1_1) {
		if (keep_alive_set == 0) {
			/* no Connection-Header sent */
			
			/* HTTP/1.1 -> keep-alive default TRUE */
			con->keep_alive = TRUE;
		}
		
		/* RFC 2616, 14.23 */
		if (con->request.http_host == NULL) {
			con->http_status = 400;
			return 0;
		}
	} else {
		if (keep_alive_set == 0) {
			/* no Connection-Header sent */
			
			/* HTTP/1.0 -> keep-alive default FALSE  */
			con->keep_alive = FALSE;
		}
	}
	
	/* check hostname field */
	if (0 != request_check_hostname(srv, con, con->request.http_host)) {
		log_error_write(srv, __FILE__, __LINE__, "sb", 
				"(notice) Bad Hostname:", con->request.http_host);
		
		con->http_status = 400;
		
		return 0;
	}
	
	/* check if we have read post data */
	if (con->request.http_method == HTTP_METHOD_POST) {
		
		if (con->request.http_content_type == NULL) {
			log_error_write(srv, __FILE__, __LINE__, "ds", 
					con->fd, "(notice) POST request, but content-type not set");
		}
		
		if (con_length_set == 0) {
			/* content-length is missing */
			
			con->http_status = 411;
			return 0;
		} 
		
		/* if this is more then just the header
		 * move the rest of the request-packet to the post-data section 
		 */
		
		if (con->request.request->used != con->header_len) {
			buffer_copy_string_len(con->request.content, 
					       con->request.request->ptr + con->header_len, 
					       con->request.request->used - con->header_len);
		}
		
		if (con->request.request->used - con->header_len == con->request.content_length) {
			/* all the post data is here */
			
		} else if (con->request.request->used - con->header_len == con->request.content_length + 2) {
			/* additional \r\n at the end (w3m) */
			
			log_error_write(srv, __FILE__, __LINE__, "ds", 
					con->fd, "(notice) POST request, additional newline");
			
		} else {
			/* have to read more data */
			
			buffer_prepare_append(con->request.content, con->request.content_length);
			
			return 1;
		}
	}
	
	return 0;
}

int http_request_header_finished(server *srv, connection *con) {
	if (con->request.request->used < 5) return 0;
	
	if (0 == memcmp(con->request.request->ptr + con->request.request->used - 5, "\r\n\r\n", 4)) return 1;
	if (NULL != strstr(con->request.request->ptr, "\r\n\r\n")) return 1;
	
	return 0;
}
