#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/resource.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <grp.h>
#include <pwd.h>
#include <assert.h>
#include <locale.h>

#include <stdio.h>

#include "server.h"
#include "buffer.h"
#include "network.h"
#include "log.h"
#include "keyvalue.h"
#include "response.h"
#include "request.h"
#include "chunk.h"
#include "http_chunk.h"
#include "fdevent.h"
#include "connections.h"
#include "file_cache.h"
#include "plugin.h"
#include "joblist.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef USE_OPENSSL
# include <openssl/ssl.h> 
# include <openssl/err.h> 
# include <openssl/rand.h> 
#endif


#ifndef __sgi
/* IRIX doesn't like the alarm based time() optimization */
/* #define USE_ALARM */
#endif

sig_atomic_t srv_shutdown = 0;
sig_atomic_t handle_sig_alarm = 1;
sig_atomic_t handle_sig_hup = 0;

#if defined(HAVE_SIGACTION) && defined(SA_SIGINFO)
void sigaction_handler(int sig, siginfo_t *si, void *context) {
	switch (sig) {
	case SIGTERM: srv_shutdown = 1; break;
	case SIGALRM: handle_sig_alarm = 1; break;
	case SIGHUP:  handle_sig_hup = 1; break;
	}
}
#else
void signal_handler(int sig) {
	switch (sig) {
	case SIGTERM: srv_shutdown = 1; break;
	case SIGALRM: handle_sig_alarm = 1; break;
	case SIGHUP:  handle_sig_hup = 1; break;
	}
}
#endif

void daemonize() {
#ifdef SIGTTOU
	signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
	signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
	signal(SIGTSTP, SIG_IGN);
#endif
	if (fork() != 0) exit(0);
	
	if (setsid() == -1) exit(0);

	signal(SIGHUP, SIG_IGN);

	if (fork() != 0) exit(0);
	
	chdir("/");
	
	umask(0);
}

server *server_init() {
	int i;
	
	server *srv = malloc(sizeof(server));
	memset(srv, 0, sizeof(server));
	
#define CLEAN(x) \
	srv->x = buffer_init();
	
	CLEAN(parse_response);
	CLEAN(response_header);
	CLEAN(parse_full_path);
	CLEAN(ts_debug_str);
	CLEAN(ts_date_str);
	CLEAN(error_log);
	CLEAN(response_range);
	CLEAN(tmp_buf);
	CLEAN(file_cache_etag);
	CLEAN(range_buf);
	
	CLEAN(srvconf.error_logfile);
	CLEAN(srvconf.groupname);
	CLEAN(srvconf.username);
	CLEAN(srvconf.changeroot);
	CLEAN(srvconf.bindhost);
	CLEAN(srvconf.errorfile_prefix);
	CLEAN(srvconf.ssl_pemfile);
	CLEAN(srvconf.license);
	CLEAN(srvconf.event_handler);
	CLEAN(srvconf.pid_file);
	
	CLEAN(conf.document_root);
	
	CLEAN(tmp_chunk_len);
#undef CLEAN
	
#define CLEAN(x) \
		x = array_init();
	
	CLEAN(srv->conf.indexfiles);
	CLEAN(srv->conf.mimetypes);
	CLEAN(srv->conf.cgi);
	CLEAN(srv->config_context);
	CLEAN(srv->config_touched);
#undef CLEAN
	
	for (i = 0; i < FILE_CACHE_MAX; i++) {
		srv->mtime_cache[i].str = buffer_init();
	}
	
	srv->cur_ts = time(NULL);
	
	srv->conns = malloc(sizeof(connections));
	assert(srv->conns != NULL);
	memset(srv->conns, 0, sizeof(connections));
	
	srv->joblist = malloc(sizeof(connections));
	assert(srv->joblist != NULL);
	memset(srv->joblist, 0, sizeof(connections));
	
	srv->file_cache = file_cache_init();
	assert(srv->file_cache != NULL);
	
	srv->srvconf.modules = array_init();
	
	/* use syslog */
	srv->log_error_fd = -1;
	
	return srv;
}

void server_free(server *srv) {
	size_t i;
	
	for (i = 0; i < FILE_CACHE_MAX; i++) {
		buffer_free(srv->mtime_cache[i].str);
	}
	
#define CLEAN(x) \
	buffer_free(srv->x);
	
	CLEAN(parse_response);
	CLEAN(response_header);
	CLEAN(parse_full_path);
	CLEAN(ts_debug_str);
	CLEAN(ts_date_str);
	CLEAN(error_log);
	CLEAN(response_range);
	CLEAN(tmp_buf);
	CLEAN(file_cache_etag);
	CLEAN(range_buf);
	
	CLEAN(srvconf.error_logfile);
	CLEAN(srvconf.groupname);
	CLEAN(srvconf.username);
	CLEAN(srvconf.changeroot);
	CLEAN(srvconf.bindhost);
	CLEAN(srvconf.ssl_pemfile);
	CLEAN(srvconf.errorfile_prefix);
	CLEAN(srvconf.license);
	CLEAN(srvconf.event_handler);
	CLEAN(srvconf.pid_file);
	
	CLEAN(conf.document_root);
	
	CLEAN(tmp_chunk_len);
#undef CLEAN
	
#define CLEAN(x) \
		array_free(x);
	
	CLEAN(srv->conf.indexfiles);
	CLEAN(srv->conf.mimetypes);
	CLEAN(srv->conf.cgi);
	CLEAN(srv->config_context);
	CLEAN(srv->config_touched);
#undef CLEAN
	
	for (i = 0; i < srv->stack.size; i++) {
		free(srv->stack.ptr[i]);
	}
	
	if (srv->stack.ptr) free(srv->stack.ptr);
	fdevent_free(srv->ev);
	
	connections_free(srv);
	free(srv->conns);
	
	joblist_free(srv->joblist);
	
	file_cache_free(srv->file_cache);
	
	array_free(srv->srvconf.modules);
	
	free(srv);
}

void show_version () {
#ifdef USE_OPENSSL
# define TEXT_SSL " (ssl)"
#else
# define TEXT_SSL
#endif
	char *b = PACKAGE "-" VERSION TEXT_SSL \
" - a light and fast webserver\n" \
"Build-Date: " __DATE__ " " __TIME__ "\n";
;
#undef TEXT_SSL	
	write(STDOUT_FILENO, b, strlen(b));
}

void show_help () {
#ifdef USE_OPENSSL
# define TEXT_SSL " (ssl)"
#else
# define TEXT_SSL
#endif
	char *b = PACKAGE "-" VERSION TEXT_SSL " ("__DATE__ " " __TIME__ ")" \
" - a light and fast webserver\n" \
"usage:\n" \
" -d <dir>   document-root\n" \
" -f <name>  filename of the config-file\n" \
" -p <port>  bind to tcp-port (default: for http: 80, for https: 443)\n" \
" -b <ip>    bind to ip (default: bind all interfaces)\n" \
" -e <name>  name of the error logfile  (default: stderr)\n" \
" -D         don't go to background (default: go to background)\n" \
" -s         use HTTPS instead of HTTP\n" \
" -v         show version\n" \
" -h         show this help\n" \
"(root only)\n" \
" -c <dir>   chroot to directory\n" \
" -u <user>  change to username\n" \
" -g <group> change to groupname\n" \
"\n"
;
#undef TEXT_SSL	
	write(STDOUT_FILENO, b, strlen(b));
}

int main (int argc, char **argv) {
	server *srv = NULL;
	int i_am_root;
	int o;
	int num_childs = 0;
	int pid_fd = -1;
	size_t i;
#ifdef HAVE_SIGACTION
	struct sigaction act;
#endif
	
#ifdef USE_ALARM
	struct itimerval interval;
	
	interval.it_interval.tv_sec = 1;
	interval.it_interval.tv_usec = 0;
	interval.it_value.tv_sec = 1;
	interval.it_value.tv_usec = 0;
#endif
	
	/* for nice %b handling in strfime() */
	setlocale(LC_TIME, "C");
	
	if (NULL == (srv = server_init())) {
		fprintf(stderr, "sigprocmask failed ... report me ... bye ...\n");
		return -1;
	}
	
	/* init structs done */
	
	srv->srvconf.port = 0;
	i_am_root = (getuid() == 0);
	srv->srvconf.dont_daemonize = 0;
	
	while(-1 != (o = getopt(argc, argv, "c:b:d:e:f:g:hp:u:vD6"))) {
		switch(o) {
		case 'f': if (config_read(srv, optarg)) { return -1;} break;
		case 'd': buffer_copy_string(srv->conf.document_root, optarg); /* docroot */ break;
		case 'p': srv->srvconf.port = strtol(optarg, NULL, 10);/* port */ break;
		case 'b': buffer_copy_string(srv->srvconf.bindhost, optarg); /* hostname */ break;
		case 'e': buffer_copy_string(srv->srvconf.error_logfile, optarg); /* logfilename */ break;
		case 'D': srv->srvconf.dont_daemonize = 1; break;
		case '6': srv->srvconf.use_ipv6 = 1; break;
		case 'v': show_version(); return 0;
		case 'h': show_help(); return 0;
		case 'c': if (i_am_root) { buffer_copy_string(srv->srvconf.changeroot, optarg); }/* chroot() */ break;
		case 'u': if (i_am_root) { buffer_copy_string(srv->srvconf.username, optarg); } /* set user */ break;
		case 'g': if (i_am_root) { buffer_copy_string(srv->srvconf.groupname, optarg); } /* set group */ break;
		default: 
			show_help();
			return -1;
		}
	}
	
	/* close stdin and stdout, as they are not needed */
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	
	if (0 != config_set_defaults(srv)) {
		log_error_write(srv, __FILE__, __LINE__, "s", 
				"setting default values failed");
		server_free(srv);
		return -1;
	}
	
	/* UID handling */
	if (!i_am_root && (geteuid() == 0 || getegid() == 0)) {
		/* we are setuid-root */
		
		log_error_write(srv, __FILE__, __LINE__, "s", 
				"Are you nuts ? Don't apply a SUID bit to this binary");
		
		server_free(srv);
		return -1;
	}
	
	/* check document-root */
	if (srv->conf.document_root->used <= 1) {
		log_error_write(srv, __FILE__, __LINE__, "s", 
				"document-root is not set\n");
		
		server_free(srv);
		
		return -1;
	}
	
	if (plugins_load(srv)) {
		log_error_write(srv, __FILE__, __LINE__, "s",
				"loading plugins finally failed");
		
		plugins_free(srv);
		server_free(srv);
		
		return -1;
	}
	
	/* open pid file BEFORE chroot */
	if (srv->srvconf.pid_file->used) {
		if (-1 == (pid_fd = open(srv->srvconf.pid_file->ptr, O_WRONLY | O_CREAT | O_EXCL | O_TRUNC, S_IRUSR | S_IWUSR))) {
			struct stat st;
			if (errno != EEXIST) {
				log_error_write(srv, __FILE__, __LINE__, "sbs",
					"opening pid-file failed:", srv->srvconf.pid_file, strerror(errno));
				return -1;
			}
			
			if (0 != stat(srv->srvconf.pid_file->ptr, &st)) {
				log_error_write(srv, __FILE__, __LINE__, "sbs",
						"stating existing pid-file failed:", srv->srvconf.pid_file, strerror(errno));
			}
			
			if (!S_ISREG(st.st_mode)) {
				log_error_write(srv, __FILE__, __LINE__, "sb",
						"pid-file exists and isn't regular file:", srv->srvconf.pid_file);
				return -1;
			}
			
			if (-1 == (pid_fd = open(srv->srvconf.pid_file->ptr, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR))) {
				log_error_write(srv, __FILE__, __LINE__, "sbs",
						"opening pid-file failed:", srv->srvconf.pid_file, strerror(errno));
				return -1;
			}
		}
	}
	
	
	if (i_am_root) {
		struct group *grp = NULL;
		struct passwd *pwd = NULL;
		struct rlimit rlim;
		
		/* set rlimits */
		if (0 != getrlimit(RLIMIT_NOFILE, &rlim)) {
			log_error_write(srv, __FILE__, __LINE__,
					"ss", "couldn't get 'max filedescriptors'",
					strerror(errno));
			return -1;
		}
		
		rlim.rlim_cur = 8192;
		rlim.rlim_max = 8192;
			
		if (0 != setrlimit(RLIMIT_NOFILE, &rlim)) {
			log_error_write(srv, __FILE__, __LINE__,
					"ss", "couldn't set 'max filedescriptors'",
					strerror(errno));
			return -1;
		}
		
		/* set user and group */
		if (srv->srvconf.username->used) {
			if (NULL == (pwd = getpwnam(srv->srvconf.username->ptr))) {
				log_error_write(srv, __FILE__, __LINE__, "sb", 
						"can't find username", srv->srvconf.username);
				return -1;
			}
			
			if (pwd->pw_uid == 0) {
				log_error_write(srv, __FILE__, __LINE__, "s",
						"I will not set uid to 0\n");
				return -1;
			}
		}
		
		if (srv->srvconf.groupname->used) {
			if (NULL == (grp = getgrnam(srv->srvconf.groupname->ptr))) {
				log_error_write(srv, __FILE__, __LINE__, "sb", 
					"can't find groupname", srv->srvconf.groupname);
				return -1;
			}
			if (grp->gr_gid == 0) {
				log_error_write(srv, __FILE__, __LINE__, "s",
						"I will not set gid to 0\n");
				return -1;
			}
		}
		
		if (0 != network_server_init(srv, srv->srvconf.bindhost->ptr, srv->srvconf.port)) {
			plugins_free(srv);
			server_free(srv);
			
			return -1;
		}
		
		if (srv->srvconf.changeroot->used) {
			if (-1 == chroot(srv->srvconf.changeroot->ptr)) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "chroot failed: ", strerror(errno));
				return -1;
			}
			if (-1 == chdir("/")) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "chdir failed: ", strerror(errno));
				return -1;
			}
		}

		/* drop root privs */
		if (srv->srvconf.groupname->used) setgid(grp->gr_gid);
		if (srv->srvconf.username->used) setuid(pwd->pw_uid);
	} else {
		if (0 != network_server_init(srv, srv->srvconf.bindhost->ptr, srv->srvconf.port)) {
			plugins_free(srv);
			server_free(srv);
			return -1;
		}
	}
	
	if (HANDLER_GO_ON != plugins_call_init(srv)) {
		log_error_write(srv, __FILE__, __LINE__, "s", "Initialization of plugins failed. Going down.");
		
		plugins_free(srv);
		server_free(srv);
		
		return -1;
	}
	
	if (HANDLER_GO_ON != plugins_call_set_defaults(srv)) {
		log_error_write(srv, __FILE__, __LINE__, "s", "Configuration of plugins failed. Going down.");
		
		plugins_free(srv);
		server_free(srv);
		
		return -1;
	}
	
	/* dump unused config-keys */
	for (i = 0; i < srv->config->used; i++) {
		data_unset *du = srv->config->data[i];
		
		if (NULL == array_get_element(srv->config_touched, du->key->ptr)) {
			log_error_write(srv, __FILE__, __LINE__, "sbs", 
					"WARNING: unknown config-key:",
					du->key,
					"(ignored)");
		}
	}
	
	if (srv->config_deprecated) {
		log_error_write(srv, __FILE__, __LINE__, "s", 
				"Configuration contains deprecated keys. Going down.");
		
		plugins_free(srv);
		server_free(srv);
		
		return -1;
	}
	
	/* network is up, let's deamonize ourself */
	if (srv->srvconf.dont_daemonize == 0) daemonize();
	
	/* write pid file */
	if (pid_fd != -1) {
		buffer_copy_long(srv->tmp_buf, getpid());
		write(pid_fd, srv->tmp_buf->ptr, srv->tmp_buf->used - 1);
		close(pid_fd);
	}
	
	if (-1 == log_error_open(srv)) {
		log_error_write(srv, __FILE__, __LINE__, "s", 
				"opening errorlog failed, dying");
		
		return -1;
	}
	
	/* init SSL */
#ifdef USE_OPENSSL
	if (srv->is_ssl) {
		SSL_load_error_strings();
		SSL_library_init();
		
		if (0 == RAND_status()) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:", 
					"not enough entropy in the pool");
			return -1;
		}
		
		if (NULL == (srv->ssl_ctx = SSL_CTX_new(SSLv23_server_method()))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:", 
					ERR_error_string(ERR_get_error(), NULL));
			return -1;
		}
		
		if (0 > SSL_CTX_use_certificate_file(srv->ssl_ctx, srv->srvconf.ssl_pemfile->ptr, SSL_FILETYPE_PEM)) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:", 
					ERR_error_string(ERR_get_error(), NULL));
			return -1;
		}	
		
		if (0 > SSL_CTX_use_PrivateKey_file (srv->ssl_ctx, srv->srvconf.ssl_pemfile->ptr, SSL_FILETYPE_PEM)) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:", 
					ERR_error_string(ERR_get_error(), NULL));
			return -1;
		}
		
		if (!SSL_CTX_check_private_key(srv->ssl_ctx)) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:", 
					"Private key does not match the certificate public key");
			return -1;
		}
	}
#endif
	
#ifdef HAVE_SIGACTION
	memset(&act, 0, sizeof(act));
	act.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &act, NULL);

# ifndef SA_SIGINFO
	act.sa_handler = signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
# else
	act.sa_sigaction = sigaction_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_SIGINFO;
# endif
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGHUP,  &act, NULL);
	
#else
	/* ignore the SIGPIPE from sendfile() */
	signal(SIGPIPE, SIG_IGN);
	signal(SIGTERM, signal_handler);
	signal(SIGHUP,  signal_handler);
#endif
		
#ifdef USE_ALARM
	signal(SIGALRM, signal_handler);
	
	/* setup periodic timer (1 second) */
	if (setitimer(ITIMER_REAL, &interval, NULL)) {
		log_error_write(srv, __FILE__, __LINE__, "setting timer failed");
		return -1;
	}
	
	getitimer(ITIMER_REAL, &interval);
#endif
	
	fdevent_reset(srv->ev);
	fdevent_event_add(srv->ev, -1, srv->fd, FDEVENT_IN);
	
	/* start watcher and workers */
	num_childs = 0;
	if (num_childs > 0) {
		int child = 0;
		while (!child && !srv_shutdown) {
			if (num_childs > 0) {
				switch (fork()) {
				case -1:
					return -1;
				case 0:
					child = 1;
					break;
				default:
					num_childs--;
					break;
				}
			} else {
				int status;
				wait(&status);
				num_childs++;
			}
		}
		if (!child) return 0;
	}
	
	/* main-loop */
	while (!srv_shutdown) {
		int n;
		size_t ndx;
		time_t min_ts;
		
		if (handle_sig_hup) {
			handler_t r;
			
			/* reset notification */
			handle_sig_hup = 0;
			
			
			/* cycle logfiles */
			
			switch(r = plugins_call_handle_sighup(srv)) {
			case HANDLER_GO_ON:
				break;
			default:
				log_error_write(srv, __FILE__, __LINE__, "sd", "sighup-handler return with an error", r);
				break;
			}
			
			if (-1 == log_error_cycle(srv)) {
				log_error_write(srv, __FILE__, __LINE__, "s", "cycling errorlog failed, dying");
				
				return -1;
			}
		}
		
		if (handle_sig_alarm) {
			/* a new second */
			
#ifdef USE_ALARM
			/* reset notification */
			handle_sig_alarm = 0;
#endif
			
			/* get current time */
			min_ts = time(NULL);
			
			if (min_ts != srv->cur_ts) {
				int cs = 0;
				connections *conns = srv->conns;
				handler_t r;
				
				switch(r = plugins_call_handle_trigger(srv)) {
				case HANDLER_GO_ON:
					break;
				default:
					log_error_write(srv, __FILE__, __LINE__, "d", r);
					break;
				}
				
				/* trigger waitpid */
				srv->cur_ts = min_ts;
				
				/**
				 * check all connections for timeouts 
				 * 
				 */
				for (ndx = 0; ndx < conns->used; ndx++) {
					int changed = 0;
					connection *con;
					
					con = conns->ptr[ndx];
						
					if ((con->state == READ) &&
					    (srv->cur_ts - con->read_idle_ts > 60)) {
						/* time - out */
#if 0
						log_error_write(srv, __FILE__, __LINE__, "sd", 
								"connection closed - read-timeout:", con->fd);
#endif
						connection_set_state(srv, con, ERROR);
						changed = 1;
					}
					
					if ((con->state == WRITE) &&
					    (con->write_request_ts != 0)) { 
#if 0
						if (srv->cur_ts - con->write_request_ts > 60) {
							log_error_write(srv, __FILE__, __LINE__, "sdd", 
									"connection closed - pre-write-request-timeout:", con->fd, srv->cur_ts - con->write_request_ts);
						}
#endif
						
						if (srv->cur_ts - con->write_request_ts > 360) {
							/* time - out */
#if 1
							log_error_write(srv, __FILE__, __LINE__, "sd", 
									"connection closed - write-request-timeout:", con->fd);
#endif
							connection_set_state(srv, con, ERROR);
							changed = 1;
						}
					}
					
					if (changed) {
						fd_conn_buffer *fdc = srv->ev->fdc;
						
						connection_state_machine(srv, con);
						
						/* if the connection is new or ... */
						if (con->fde_ndx == -1 || 
						    fdc->ptr[con->fde_ndx].fd_type == FD_EVENT_TYPE_CONNECTION) {
							switch(con->state) {
							case READ_POST:
							case READ:
							case CLOSE:
								con->fde_ndx = fdevent_event_add(srv->ev, con->fde_ndx, con->fd, FDEVENT_IN);
								break;
							case WRITE:
								con->fde_ndx = fdevent_event_add(srv->ev, con->fde_ndx, con->fd, FDEVENT_OUT);
								break;
							default:
								con->fde_ndx = fdevent_event_del(srv->ev, con->fde_ndx, con->fd);
								break;
							}
						}
					}
					
#if 0
					if (cs == 0) {
						fprintf(stderr, "connection-state: ");
						cs = 1;
					}
					
					fprintf(stderr, "c[%d,%d]: %s ",
						con->fd,
						con->fcgi.fd,
						connection_get_state(con->state));
#endif
				}
				
				if (cs == 1) fprintf(stderr, "\n");
			}
		}

		if ((n = fdevent_poll(srv->ev, 1000)) > 0) {

			/* n is the number of events */
			int revents;
			size_t fd_ndx;
			
			int handle_server_socket = 0;
			int handle_server_socket_revents = 0;
			
			fd_conn_buffer *fdc = srv->ev->fdc;
			
			for (fd_ndx = fdevent_event_next_fdndx(srv->ev, -1); n > 0; n--, fd_ndx = fdevent_event_next_fdndx(srv->ev, fd_ndx)) {
				int fd;
				i = 0;
				
				revents = fdevent_event_get_revent(srv->ev, fd_ndx);
				fd = fdevent_event_get_fd(srv->ev, fd_ndx);
#ifdef USE_POLL
				if (revents == POLLNVAL) {
					/* should never happen */
					SEGFAULT();
				}
#endif
				if (srv->fd == fd) {
					/* handle new requests AFTER the current requests have been processed */
					handle_server_socket = 1;
					handle_server_socket_revents = revents;
				} else {
					if (-1 == (i = fdevent_fdc_get_index(srv->ev->fdc, fd))) {
#if 1
						log_error_write(srv, __FILE__, __LINE__, "sddd", "fdevent_fd not found:", fd, fdc->used, revents);
#endif
					} else {
						connection *con = NULL;
						file_cache_entry *fce = NULL;
						int r;
						
						fdc->ptr[i].revents = revents;
						
						/* transform fd to poll-ndx */

						switch (fdc->ptr[i].fd_type) {
						case FD_EVENT_TYPE_CONNECTION:
							con = fdc->ptr[i].conn;
#if 0							
							fprintf(stderr, "%s.%d: %d: _fd(%d)_, cgi(%d), fcgi(%d) revents(%d) mode(%d)\n", 
								__FILE__, __LINE__,
								i, con->fd, con->cgi.fd, 
								con->fcgi.fd,
								revents,
								con->mode);
#endif
							connection_handle_fdevent(srv, con, revents);
							
							joblist_append(srv, con);
							
							break;
						case FD_EVENT_TYPE_DIRWATCH:
							fce = fdc->ptr[i].conn;
							file_cache_handle_fdevent(srv, fce, revents);
							
							break;
						default:
							con = fdc->ptr[i].conn;
							
							switch(r = plugins_call_handle_fdevent(srv, &(fdc->ptr[i]))) {
							case HANDLER_FINISHED:
							case HANDLER_GO_ON:
							case HANDLER_WAIT_FOR_EVENT:
								break;
							case HANDLER_ERROR:
								assert(con);
								
								con->mode = DIRECT;
								con->http_status = 500;
								
								connection_set_state(srv, con, HANDLE_REQUEST);
								
								break;
								
							default:
								log_error_write(srv, __FILE__, __LINE__, "d", r);
								break;
							}
							
							break;
						}
						
						

						if (con) {
#if 0
							fprintf(stderr, "%s.%d: %d: fd(%d), fcgi(%d), cgi(%d) revents(%d)\n", 
								__FILE__, __LINE__,
								i, con->fd, con->fcgi.fd, con->cgi.fd, revents);
#endif
							
						}
					}
				}
			}
			
			/* handle fdevents from the server-socket */
			if (handle_server_socket) {
				connection *con;
				
				revents = handle_server_socket_revents;
				
				if (revents != FDEVENT_IN) {
					log_error_write(srv, __FILE__, __LINE__, "sdddd", 
							"strange event for server socket",
							srv->fd,
							fd_ndx,
							revents,
							n);
					return -1;
				}
				
				while (NULL != (con = connection_accept(srv))) {
					handler_t r;
					connection_state_machine(srv, con);
					
					/* first contact */
					switch(con->state) {
					case READ_POST:
					case READ:
					case CLOSE:
						con->fde_ndx = fdevent_event_add(srv->ev, -1, con->fd, FDEVENT_IN);
						break;
					case WRITE:
						con->fde_ndx = fdevent_event_add(srv->ev, -1, con->fd, FDEVENT_OUT);
						break;
					default:
						break;
					}
					
					switch(r = plugins_call_handle_joblist(srv, con)) {
					case HANDLER_FINISHED:
					case HANDLER_GO_ON:
						break;
					default:
						log_error_write(srv, __FILE__, __LINE__, "d", r);
						break;
					}
				}
			}
			
		} else if (n < 0 && errno != EINTR) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "fdevent_poll failed:", strerror(errno));
		}
		
		for (ndx = 0; ndx < srv->joblist->used; ndx++) {
			connection *con = srv->joblist->ptr[ndx];
			fd_conn_buffer *fdc = srv->ev->fdc;
			handler_t r;
			
			connection_state_machine(srv, con);
						
			if (con->fde_ndx == -1 || 
			    fdc->ptr[con->fd].fd_type == FD_EVENT_TYPE_CONNECTION) {
				switch(con->state) {
				case READ_POST:
				case READ:
				case CLOSE:
					con->fde_ndx = fdevent_event_add(srv->ev, con->fde_ndx, con->fd, FDEVENT_IN);
					break;
				case WRITE:
					/* request write-fdevent only if we really need it 
					 * - if we have data to write
					 * - if the socket is not writable yet 
					 */
					if (!chunkqueue_is_empty(con->write_queue) && 
					    (con->is_writable == 0)) {
						con->fde_ndx = fdevent_event_add(srv->ev, con->fde_ndx, con->fd, FDEVENT_OUT);
					} else {
						con->fde_ndx = fdevent_event_del(srv->ev, con->fde_ndx, con->fd);
					}
					break;
				default:
					con->fde_ndx = fdevent_event_del(srv->ev, con->fde_ndx, con->fd);
					break;
				}
			}
			
			
			switch(r = plugins_call_handle_joblist(srv, con)) {
			case HANDLER_FINISHED:
			case HANDLER_GO_ON:
				break;
			default:
				log_error_write(srv, __FILE__, __LINE__, "d", r);
				break;
			}
			
			con->in_joblist = 0;
		}
		
		srv->joblist->used = 0;

#if 0
		fprintf(stderr, "-> (%d, %d, %d/%d/%d/%d, %ld)\n", 
			n, max_pollfds,
			srv->con_opened, 
			srv->con_read,
			srv->con_written,
			srv->con_closed, 
			srv->cur_ts - min_ts
			);
#endif
	}
	
	/* clean-up */
	log_error_close(srv);
	
	network_server_close(srv);
	
	plugins_free(srv);
	
	server_free(srv);
	
	return 0;
}
